package com.sun.xml.internal.bind.v2.model.annotation;

import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.lang.annotation.Annotation;
import java.util.Map;
import java.lang.reflect.Method;
import java.lang.reflect.Field;
import java.lang.reflect.Type;

public final class RuntimeInlineAnnotationReader extends AbstractInlineAnnotationReaderImpl<Type, Class, Field, Method> implements RuntimeAnnotationReader
{
    private final Map<Class<? extends Annotation>, Map<Package, Annotation>> packageCache;
    
    public RuntimeInlineAnnotationReader() {
        this.packageCache = new HashMap();
    }
    
    @Override
    public <A extends Annotation> A getFieldAnnotation(Class<A> annotation, Field field, Locatable srcPos) {
        return LocatableAnnotation.create(field.getAnnotation(annotation), srcPos);
    }
    
    @Override
    public boolean hasFieldAnnotation(Class<? extends Annotation> annotationType, Field field) {
        return field.isAnnotationPresent(annotationType);
    }
    
    @Override
    public boolean hasClassAnnotation(Class clazz, Class<? extends Annotation> annotationType) {
        return clazz.isAnnotationPresent(annotationType);
    }
    
    @Override
    public Annotation[] getAllFieldAnnotations(Field field, Locatable srcPos) {
        Annotation[] r = field.getAnnotations();
        for (int i = 0; i < r.length; ++i) {
            r[i] = LocatableAnnotation.create(r[i], srcPos);
        }
        return r;
    }
    
    @Override
    public <A extends Annotation> A getMethodAnnotation(Class<A> annotation, Method method, Locatable srcPos) {
        return LocatableAnnotation.create(method.getAnnotation(annotation), srcPos);
    }
    
    @Override
    public boolean hasMethodAnnotation(Class<? extends Annotation> annotation, Method method) {
        return method.isAnnotationPresent(annotation);
    }
    
    @Override
    public Annotation[] getAllMethodAnnotations(Method method, Locatable srcPos) {
        Annotation[] r = method.getAnnotations();
        for (int i = 0; i < r.length; ++i) {
            r[i] = LocatableAnnotation.create(r[i], srcPos);
        }
        return r;
    }
    
    @Override
    public <A extends Annotation> A getMethodParameterAnnotation(Class<A> annotation, Method method, int paramIndex, Locatable srcPos) {
        Annotation[] pa = method.getParameterAnnotations()[paramIndex];
        int length = pa.length;
        int i = 0;
        while (i < length) {
            Annotation a = pa[i];
            if (a.annotationType() == annotation) {
                return LocatableAnnotation.create(a, srcPos);
            }
            ++i;
        }
        return null;
    }
    
    @Override
    public <A extends Annotation> A getClassAnnotation(Class<A> a, Class clazz, Locatable srcPos) {
        return LocatableAnnotation.create(clazz.getAnnotation(a), srcPos);
    }
    
    @Override
    public <A extends Annotation> A getPackageAnnotation(Class<A> a, Class clazz, Locatable srcPos) {
        Package p = clazz.getPackage();
        if (p == null) {
            return null;
        }
        Map<Package, Annotation> cache = (Map)this.packageCache.get(a);
        if (cache == null) {
            cache = new HashMap();
            this.packageCache.put(a, cache);
        }
        if (cache.containsKey(p)) {
            return (A)cache.get(p);
        }
        A ann = LocatableAnnotation.create(p.getAnnotation(a), srcPos);
        cache.put(p, ann);
        return ann;
    }
    
    @Override
    public Class getClassValue(Annotation a, String name) {
        try {
            return (Class)a.annotationType().getMethod(name, new Class[0]).invoke(a, new Object[0]);
        }
        catch (IllegalAccessException e) {
            throw new IllegalAccessError(e.getMessage());
        }
        catch (InvocationTargetException e2) {
            throw new InternalError(Messages.CLASS_NOT_FOUND.format(new Object[] { a.annotationType(), e2.getMessage() }));
        }
        catch (NoSuchMethodException e3) {
            throw new NoSuchMethodError(e3.getMessage());
        }
    }
    
    @Override
    public Class[] getClassArrayValue(Annotation a, String name) {
        try {
            return (Class[])(Class[])a.annotationType().getMethod(name, new Class[0]).invoke(a, new Object[0]);
        }
        catch (IllegalAccessException e) {
            throw new IllegalAccessError(e.getMessage());
        }
        catch (InvocationTargetException e2) {
            throw new InternalError(e2.getMessage());
        }
        catch (NoSuchMethodException e3) {
            throw new NoSuchMethodError(e3.getMessage());
        }
    }
    
    @Override
    protected String fullName(Method m) {
        return m.getDeclaringClass().getName() + 35 + m.getName();
    }
}
